"use strict";

var _slicedToArray = function () { function sliceIterator(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"]) _i["return"](); } finally { if (_d) throw _e; } } return _arr; } return function (arr, i) { if (Array.isArray(arr)) { return arr; } else if (Symbol.iterator in Object(arr)) { return sliceIterator(arr, i); } else { throw new TypeError("Invalid attempt to destructure non-iterable instance"); } }; }();

module.exports = function minifyProperties(_ref) {
  var t = _ref.types;

  var count = new Map();
  var paths = new Map();

  function increment(name) {
    if (count.has(name)) {
      count.set(name, count.get(name) + 1);
    } else {
      count.set(name, 1);
    }
  }

  function addPath(path) {
    var name = path.node.name;

    if (paths.has(name)) {
      paths.get(name).push(path);
    } else {
      paths.set(name, [path]);
    }
  }

  var collectVisitor = {
    MemberExpression: function MemberExpression(path) {
      var property = path.get("property");
      if (property.isIdentifier() && path.node.computed === false) {
        increment(property.node.name);
        addPath(property);
      }
    },
    "ObjectProperty|ObjectMethod|ClassMethod": function ObjectPropertyObjectMethodClassMethod(path) {
      var key = path.get("key");
      if (key.isIdentifier() && path.node.computed === false) {
        increment(key.node.name);
        addPath(key);
      }
    }
  };

  /**
   * when we make a property to a computed property, we create a
   * new variable in the scope and we have to account for that. Here
   * we assume that a new var creation takes 7 characters more
   * - "var ab=" - we allocate 2 characters per variable because the
   * current mangler is not able to reuse variables in different scopes.
   *
   * We also account for quotes(2) and the semicolon(1)
   *
   * And we account for the usage
   * - 3 characters for MemberExpression properties - [ab] minus "."
   * - 4 characters for ObjectProperty, ObjectMethod, ClassMethod - [ab]
   */
  function filterNames() {
    var changes = new Map();
    var _iteratorNormalCompletion = true;
    var _didIteratorError = false;
    var _iteratorError = undefined;

    try {
      for (var _iterator = count[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
        var _step$value = _slicedToArray(_step.value, 2),
            name = _step$value[0],
            times = _step$value[1];

        var oldCost = name.length * times;
        var newCost = 7 /*declaration*/ + name.length + 2 /*quotes*/ + 1 /*semicolon*/;

        if (!paths.has(name)) {
          throw new Error("Forgot to addPath(path)?");
        }

        var pathArray = paths.get(name);

        pathArray.forEach(function (path) {
          var parent = path.parentPath;
          if (parent.isMemberExpression()) {
            newCost += 3;
          } else if (parent.isObjectProperty() || parent.isObjectMethod() || parent.isClassMethod()) {
            newCost += 4;
          } else {
            throw new Error("Unhandled type:" + parent.type);
          }
        });

        if (newCost < oldCost) {
          changes.set(name, pathArray);
        }
      }
    } catch (err) {
      _didIteratorError = true;
      _iteratorError = err;
    } finally {
      try {
        if (!_iteratorNormalCompletion && _iterator.return) {
          _iterator.return();
        }
      } finally {
        if (_didIteratorError) {
          throw _iteratorError;
        }
      }
    }

    return changes;
  }

  function updateItems(toChange, program) {
    var _iteratorNormalCompletion2 = true;
    var _didIteratorError2 = false;
    var _iteratorError2 = undefined;

    try {
      var _loop = function _loop() {
        var _step2$value = _slicedToArray(_step2.value, 2),
            name = _step2$value[0],
            paths = _step2$value[1];

        var varName = program.scope.generateUidIdentifier();

        program.get("body")[0].insertBefore([t.variableDeclaration("const", [t.variableDeclarator(varName, t.stringLiteral(name))])]);

        paths.forEach(function (path) {
          path.parentPath.node.computed = true;
          path.node.name = varName.name;
        });
      };

      for (var _iterator2 = toChange[Symbol.iterator](), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
        _loop();
      }
    } catch (err) {
      _didIteratorError2 = true;
      _iteratorError2 = err;
    } finally {
      try {
        if (!_iteratorNormalCompletion2 && _iterator2.return) {
          _iterator2.return();
        }
      } finally {
        if (_didIteratorError2) {
          throw _iteratorError2;
        }
      }
    }
  }

  return {
    name: "minify-properties",
    visitor: {
      Program: function Program(path) {
        path.traverse(collectVisitor);
        var toChange = filterNames();
        updateItems(toChange, path);
      }
    }
  };
};